/* reading and parsing configuration file */
/* $Id: readcfg.c,v 1.4 1999/02/05 06:55:45 vasim Exp $ */

#include <sys/types.h>
#include <ctype.h>
#include "readcfg.h"
#include <errno.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <regex.h>

#define MAX_LINE_LENGTH 128

char error_buffer[256];
char *config_error;

/* Read and parse config file */
/* Returns number of config lines, if no error */
/* Return 0, if error (error text in config_error) */
int read_config(char *filename, conf_rec params[], int number, char *section) {
  regex_t f_no, f_yes, f_decdigs, f_comment;
  regex_t f_hexdigs, f_string_nc, f_string, f_string_wq;
  regex_t exp[MAX_CONF_PARAM_NUM];
  regex_t hostsect;
  regmatch_t match[1];
  int found[MAX_CONF_PARAM_NUM];
  int allocated[MAX_CONF_PARAM_NUM];
  int n, curline, error, proceed;
  FILE *in;
  char line[MAX_LINE_LENGTH];
  char shostsect[MAX_LINE_LENGTH+10];
  int ignore;
  char *temp;
  char *beg;
  char *end;
  char *error_text;
  char *p;
  int escaped;

  config_error = error_buffer;

  in = fopen(filename, "r");
  if (in == NULL) {
    snprintf(error_buffer, sizeof(error_buffer),
	     "Cannot open config file '%s': %s", filename, strerror(errno));
    return 0;
  }

  /* Regular expressions compilation */
  regcomp(&f_no, F_NO, REG_EXTENDED | REG_ICASE);
  regcomp(&f_yes, F_YES, REG_EXTENDED | REG_ICASE);
  regcomp(&f_decdigs, F_DEC, REG_EXTENDED | REG_ICASE);
  regcomp(&f_hexdigs, F_HEX, REG_EXTENDED | REG_ICASE);
  regcomp(&f_string_nc, F_STRING_NC, REG_EXTENDED | REG_ICASE);
  regcomp(&f_string, F_STRING, REG_EXTENDED | REG_ICASE);
  regcomp(&f_string_wq, F_STRING_WQ, REG_EXTENDED | REG_ICASE);
  regcomp(&f_comment, F_COMMENT, REG_EXTENDED | REG_ICASE);
  for (n = 0; n < number; n++) {
    regcomp(&exp[n], params[n].param, REG_EXTENDED | REG_ICASE);
    found[n] = 0;
    allocated[n] = 0;
  }
  if (!section) {
    gethostname(line, sizeof(line));
    snprintf(shostsect, sizeof(shostsect), "^\\[%s\\]", line);
  } else
    snprintf(shostsect, sizeof(shostsect), "^\\[%s\\]", section);
  regcomp(&hostsect, shostsect, REG_EXTENDED | REG_ICASE);

  /* Parse */
  curline = 0;
  error = 0;
  ignore = 0;
  while ((fgets(line, MAX_LINE_LENGTH, in) != NULL) && !error) {
    curline++;
    proceed = 0;
    if (line[0] == '[') {
      if (!regexec(&hostsect, line, 1, match, 0))
	ignore = 0;
      else
	ignore = 1;
    } else
      if (!ignore && regexec(&f_comment, line, 1, match, 0)) {
	/* Line is not comment. Will search parameter */
	for (n = 0; n < number; n++) {
	  if (!regexec(&exp[n], line, 1, match, 0)) {
	    /* Parameter found. Parse argument */
	    if (params[n].flag_default)
	      found[n] = 1;
	    proceed = 1;
	    temp = line + match[0].rm_eo;
	    switch (params[n].type) {
	      /* Boolean parameter (without argument) */
	    case BOOL_NOP:
	      if (!params[n].value_func(params[n].addr, params[n].len,
					1, (void *) 1)) {
		error = 1;
		error_text = "Cannot set parameter value";
	      }
	      break;
	      /* Boolean parameter (with argument) */
	    case BOOL_PAR:
	      if (!regexec(&f_yes, temp, 1, match, 0)) {
		if (!params[n].value_func(params[n].addr,
					  params[n].len, 1, (void *) 1)) {
		  error = 1;
		  error_text = "Cannot set parameter value";
		}
	      } else
		if (!regexec(&f_no, temp, 1, match, 0)) {
		  if (!params[n].value_func(params[n].addr,
					    params[n].len, 1, (void *) 0)) {
		    error = 1;
		    error_text = "Cannot set parameter value";
		  }
		} else {
		  error = 1;
		  error_text = "Value must be 'On' ('Yes') or 'Off' ('No')";
		}
	      break;
	      /* String in quotas */
	    case STRING_WQ:
	      if (!regexec(&f_string_wq, temp, 1, match, 0)) {
		beg = temp + match[0].rm_so + 1;
		end = temp + match[0].rm_eo - 1;
		*end = '\0';
		if (allocated[n] && params[n].addr
		    && *((char **) params[n].addr)) {
		  free(*((char **) params[n].addr));
		  allocated[n] = 0;
		}
		if (beg == end)
		  beg = NULL;
		else {
		  allocated[n] = 1;
		  /* Look for escaped chars */
		  p = beg;
		  escaped = 0;
		  for (temp = beg; *temp; temp++) {
		    if (escaped) {
		      switch (*temp) {
		      case 'r':
			*p++ = '\r';
			break;
		      case 'n':
			*p++ = '\n';
			break;
		      case '\\':
			*p++ = '\\';
			break;
		      case 't':
			*p++ = '\t';
			break;
		      default:
			*p++ = *temp;
			break;
		      }
		      escaped = 0;
		    } else {
		      if (*temp != '\\')
			*p++ = *temp;
		      else
			escaped = 1;
		    }
		  }
		  *p = '\0';
		}
		if (!params[n].value_func(params[n].addr, params[n].len,
					  1, (void *) beg)) {
		  error = 1;
		  error_text = "Cannot set parameter value";
		}
	      } else {
		error = 1;
		error_text = "Bad string value";
	      }
	      break;
	      /* String parameters (with and without comments) */
	    case STRING_NC:
	    case STRING:
	      if (((params[n].type == STRING_NC)
		   && !regexec(&f_string_nc, temp, 1, match, 0)) ||
		  ((params[n].type == STRING)
		   && !regexec(&f_string, temp, 1, match, 0))) {
		beg = temp + match[0].rm_so;
		end = temp + match[0].rm_eo;
		*end = '\0';
		if (allocated[n] && params[n].addr
		    && *((char **) params[n].addr))
		  free(*((char **) params[n].addr));
		else
		  allocated[n] = 1;
		if (!params[n].value_func(params[n].addr, params[n].len,
					  1, (void *) beg)) {
		  error = 1;
		  error_text = "Cannot set parameter value";
		}
	      } else {
		error = 1;
		error_text = "Bad string value";
	      }
	      break;
	      /* Decimal number */
	    case DEC_DIGITS:
	      if (!regexec(&f_decdigs, temp, 1, match, 0)) {
		*(temp + match[0].rm_eo) = '\0';
		if (!params[n].value_func(params[n].addr, params[n].len, 1,
					  (void *) atoi(temp + match[0].rm_so))) {
		  error = 1;
		  error_text = "Cannot set parameter value";
		}
	      } else {
		error = 1;
		error_text = "Bad decimal value";
	      }
	      break;
	      /* Hexadecimal number */
	    case HEX_DIGITS:
	      if (!regexec(&f_hexdigs, temp, 1, match, 0)) {
		*(temp + match[0].rm_eo) = '\0';
		for (beg = temp + match[0].rm_so; *beg; beg++)
		  *beg = toupper(*beg);
		if (!params[n].value_func(params[n].addr, params[n].len, 1,
					  (void *) strtol(temp + match[0].rm_so, NULL, 16))) {
		  error = 1;
		  error_text = "Cannot set parameter value";
		}
	      } else {
		error = 1;
		error_text = "Bad hexadecimal value";
	      }
	      break;
	    case IGNORE:
	      break;
	    }
	  }
	  if (error)
	    break;
	}
	/* Any parameters not found for this line */
	if (!proceed) {
	  error = 1;
	  error_text = "Unknown parameter";
	}
      }
  }
  if (error) {
    snprintf(error_buffer, sizeof(error_buffer),
	     "Error at line %d in file '%s': %s", curline,
	     filename, error_text);
  } else for (n = 0; n < number; n++)
    if (!found[n] && params[n].flag_default && (params[n].type != IGNORE))
      if (!params[n].value_func(params[n].addr,
				params[n].len, 0, params[n].def)) {
	snprintf(error_buffer, sizeof(error_buffer),
		 "Error in file '%s': Cannot set default value for %d",
		 filename, n);
	error = 1;
      }
  /* Freeing some memory */
  regfree(&hostsect);
  for (n = number-1; n >= 0; n--)
    regfree(&exp[n]);
  regfree(&f_comment);
  regfree(&f_string_wq);
  regfree(&f_string);
  regfree(&f_string_nc);
  regfree(&f_hexdigs);
  regfree(&f_decdigs);
  regfree(&f_yes);
  regfree(&f_no);
  if (error)
    return 0;
  else
    return curline;
}

/* Set number value. addr - int pointer, value - int */
int std_DIGITS(void *addr, int len, int flag, void *value) {
  *((int *) addr) = (int) value;
  return 1;
}

/* Set boolean value. addr - int pointer, value - int */
int std_BOOL(void *addr, int len, int flag, void *value) {
  *((int *) addr) = (int) value;
  return 1;
}

/* Set string value (possible memory allocate) */
int std_STRING(void *addr, int len, int flag, void *value) {
  if (value) {
    if (len) {
      strncpy(*((char **) addr), (char *) value, len-1);
      *(*((char **) addr) + len - 1) = '\0';
    } else {
      *((char **) addr) = malloc(strlen((char *) value) + 1);
      if (!*((char **) addr))
	return 0;
      strcpy(*((char **) addr), (char *) value);
    }
  } else
    *((char **) addr) = NULL;
  return 1;
}
